#include <stdlib.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <jpeglib.h>
#include <png.h>

static int
int_load_jpeg (const char     *path,
           int            *width,
           int            *height,
           int            *row_stride,
           unsigned char **pixels)
{
  struct jpeg_decompress_struct  cinfo;
  struct jpeg_error_mgr          jerr;
  FILE                          *infile;
  JSAMPARRAY                     buffer;
  int row=0;

  if ((infile = fopen (path, "r")) == NULL)
    {
      fprintf (stderr, "can't open %s\n", path);
      return -1;
    }

  jpeg_create_decompress (&cinfo);
  cinfo.err = jpeg_std_error (&jerr);
  jpeg_stdio_src (&cinfo, infile);

  (void) jpeg_read_header (&cinfo, TRUE);
  (void) jpeg_start_decompress (&cinfo);

  if (cinfo.output_components != 3)
    {
      fprintf (stderr, "\n\nerror loading JPEG %s does not have 3 components\n\n", path);
      exit (-1);
      return -1;
    }

  *row_stride = cinfo.output_width * cinfo.output_components;

  if ((*row_stride) % 2)
    (*row_stride)++;

  buffer = (*cinfo.mem->alloc_sarray)
    ((j_common_ptr) &cinfo, JPOOL_IMAGE, *row_stride, 1);

  /* allocate memory for the bounding image that is a multiple of 2 */ 
  *pixels = calloc (*row_stride * cinfo.output_height+ (cinfo.output_width%2)  ,1);
 
  while (cinfo.output_scanline < cinfo.output_height)
    {
      (void) jpeg_read_scanlines (&cinfo, buffer, 1);
      memcpy ((*pixels) + (*row_stride) * (row++), buffer[0], *row_stride);
    }
  *width  = cinfo.output_width;
  *height = cinfo.output_height;
  jpeg_destroy_decompress (&cinfo);
  fclose (infile);
  return 0;
}


static int
load_png (const char     *path,
          int            *width,
          int            *height,
          int            *row_stride,
          unsigned char **pixels)
{
  png_uint_32 w;
  png_uint_32 h;
  FILE         *infile;
  png_structp   load_png_ptr;
  png_infop     load_info_ptr;
  unsigned char header[8];

  unsigned    int i;
  png_bytep  *row_p = NULL;

  infile = fopen (path, "rb");
  if (!infile)
    {
      return -1;
    }

  fread (header, 1, 8, infile);
  if (png_sig_cmp (header, 0, 8))
    {
      fclose (infile);
      printf ("not png\n");
      return -1;
    }

  load_png_ptr = 
    png_create_read_struct (PNG_LIBPNG_VER_STRING, NULL, NULL, NULL);
  if (!load_png_ptr)
    {
      fclose (infile);
      return -1;
    }

  load_info_ptr = png_create_info_struct (load_png_ptr);
  if (!load_info_ptr)
    {
      png_destroy_read_struct (&load_png_ptr, &load_info_ptr, NULL);
      fclose (infile);
      return -1;
    }

  if (setjmp (png_jmpbuf (load_png_ptr)))
    {
      png_destroy_read_struct (&load_png_ptr, &load_info_ptr, NULL);
      if (row_p)
        free (row_p);
      fclose (infile);
      return -1;
    }

  png_init_io (load_png_ptr, infile);
  png_set_sig_bytes (load_png_ptr, 8);
  png_read_info (load_png_ptr, load_info_ptr);
  {
    int bit_depth;
    int color_type;

    png_get_IHDR (load_png_ptr, load_info_ptr, &w, &h, &bit_depth, &color_type, NULL, NULL, NULL);
    *width = w;
    *height = h;
    if (bit_depth != 8 || !(color_type == PNG_COLOR_TYPE_RGB || color_type == PNG_COLOR_TYPE_RGBA))
      {
        printf ("bpp or type mismatch\n");
        png_destroy_read_struct (&load_png_ptr, &load_info_ptr, NULL);
        fclose (infile);
        return -1;
      }
    if (color_type == PNG_COLOR_TYPE_RGBA)
      png_set_strip_alpha (load_png_ptr);
  }

    *row_stride = w * 3;
    *pixels = calloc (*row_stride * h,1);

    row_p = (png_bytep *) malloc (sizeof (png_bytep) * h);

    for (i=0; i< h; i++)
      row_p[i] = *pixels + *row_stride * i;

    png_read_image (load_png_ptr, row_p);
    png_read_end (load_png_ptr, NULL);

    png_destroy_read_struct (&load_png_ptr, &load_info_ptr, NULL);

    if (row_p)
      free (row_p);
    fclose (infile);
  
  return 0;
}

int
load_image (const char     *path,
            int            *width,
            int            *height,
            int            *row_stride,
            unsigned char **pixels)
{
  if (strstr (path, ".jpg")  ||
      strstr (path, ".jpeg") ||
      strstr (path, ".JPG")  ||
      strstr (path, ".JPEG")
   )
    return int_load_jpeg (path, width, height, row_stride, pixels);
  if (strstr (path, ".png") ||
      strstr (path, ".PNG"))
    return load_png (path, width, height, row_stride, pixels);
  return -1;
}


/* do factor 2 decimation on the image data
 *
 * if some initial cropping is to be done, do it here
 */
void
downsize_to (int min_dim,
             int *width,
             int *height,
             int *row_stride,
             unsigned char **pixels)
{
  /* should realloc the original image, if it doesn't contained padded space
   * (as well as pad it with proper "expand" color
   */
  while (*width  /2  > min_dim &&
         *height /2  > min_dim)
  {
      int            width2, height2, row_stride2;
      unsigned char *pixels2;
      int            x,y;
      width2  = *width/2;
      height2 = *height/2;
      row_stride2 = width2*3;
      if (row_stride2 % 2)
        row_stride2++;
      pixels2 = malloc (row_stride2 * height2 + (height2%2));
      
      for (y=0;y<height2;y++)
        {
           unsigned char *src00 = &((*pixels)[(*row_stride) * (y*2)]);
           unsigned char *src10 = &((*pixels)[(*row_stride) * (y*2)+3]);
           unsigned char *src01 = &((*pixels)[(*row_stride) * (y*2+1)]);
           unsigned char *src11 = &((*pixels)[(*row_stride) * (y*2+1)+3]);
           unsigned char *dst   = &pixels2[row_stride2 * y];
           for (x=0;x<width2;x++)
             {
                int c;
                for (c=0;c<3;c++)
                  {
                    *dst = (*src00 + *src10 + *src01 + *src11) / 4;
                    dst++;
                    src00++;
                    src10++;
                    src01++;
                    src11++;
                  }
             }
        }
      free (*pixels);
      *pixels = pixels2;
      *width = width2;
      *height = height2;
      *row_stride = row_stride2;
  }
}


